# Generated by Django 5.0.4 on 2024-05-08 12:34

import django.utils.timezone
import r_shepard.api.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("api", "0016_project_commit_interval_project_last_commit_time_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="container",
            name="container_id",
            field=models.CharField(
                default=r_shepard.api.models.Container.generate_container_id,
                help_text="The unique identifier of the container within the project.",
                max_length=8,
                verbose_name="Container ID",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="cpus",
            field=models.PositiveIntegerField(
                default=2,
                help_text="The maximum number of CPU cores the container can use. Will only take effect after restarting the container.",
                verbose_name="CPU Allocation",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="is_running",
            field=models.BooleanField(
                default=False,
                help_text="Indicates whether the container is running",
                verbose_name="Running",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="local_url",
            field=models.URLField(
                default="",
                help_text="Local URL to access this container's RStudio instance",
                max_length=1024,
                verbose_name="Local URL",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="port",
            field=models.PositiveIntegerField(
                default=8787,
                help_text="The host port on which the container's RStudio instance is accessible locally.",
                verbose_name="Host Port",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="ram",
            field=models.PositiveIntegerField(
                default=2,
                help_text="The maximum amount of RAM in GB the container can use. Will only take effect after restarting the container.",
                verbose_name="RAM Allocation (GB)",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="tailscale_funnel_url",
            field=models.URLField(
                default="",
                help_text="Tailscale Funnel URL to access this container's RStudio instance from the web",
                max_length=1024,
                verbose_name="Tailscale Funnel URL",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="tailscale_serve_url",
            field=models.URLField(
                default="",
                help_text="Tailscale Serve URL to access this container's RStudio instance from the tailnet",
                max_length=1024,
                null=True,
                verbose_name="Tailscale Serve URL",
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="commit_interval",
            field=models.PositiveIntegerField(
                default=15, help_text="Commit interval in minutes."
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="last_commit_time",
            field=models.DateTimeField(
                default=django.utils.timezone.now,
                help_text="Last time the project was committed to the Git repository.",
            ),
        ),
    ]



from django.conf import settings
from podman import PodmanClient
from podman.errors import NotFound

from .models import Container, Project

# Get the podman settings from the Django settings
PODMAN_HOST_ADDRESS = settings.PODMAN_HOST_ADDRESS
PODMAN_SOCKET = settings.PODMAN_SOCKET


# 'container' always refers to the Container model instance whereas
# 'podman_container' refers to the podman container instance


class PodmanError(Exception):
    """Base class for Podman exceptions."""

    pass


def get_podman_container(client, project, container):
    """Get the podman container if it exists, else return None."""
    try:
        return client.containers.get(f"rstudio_{project.slug}_{container.container_id}")
    except NotFound:
        return None


def start_and_save_container(podman_container, container):
    """Start the podman container and save the container details."""
    podman_container.start()
    podman_container.reload()

    container.port = podman_container.ports["8787/tcp"][0]["HostPort"]
    container.local_url = f"http://{PODMAN_HOST_ADDRESS}:{container.port}"
    container.save()

    print(f"Container started at {container.local_url}")


def create_container_config(project, container, password):
    """Create the container configuration. Right now, it only supports RStudio."""
    return {
        "image": "docker.io/rocker/rstudio",
        "name": f"rstudio_{project.slug}_{container.container_id}",
        "detach": True,
        "ports": {"8787/tcp": container.port},  # Assuming default RStudio port
        "environment": {
            "PASSWORD": password,  # Set the PASSWORD environment variable for RStudio
            "ROOT": "false",
        },
        "cpu_count": container.cpus,
        "mem_limit": f"{container.ram}g",
        "mounts": [
            {
                "type": "bind",
                "source": f"{project.workspace_path}",
                "target": f"/root/{project.slug}",
                "read_only": False,
            },
            {
                "type": "bind",
                "source": f"{project.data_path}",
                "target": "/data",
                "read_only": True,
            },
        ],
    }


def start_container(project: Project, container: Container, password: str = None):
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        podman_container = get_podman_container(client, project, container)

        if podman_container is not None:
            print("Container already exists. No need to create.")
            if podman_container.status != "running":
                print("Starting container...")
                start_and_save_container(podman_container, container)
        else:
            print("Container does not exist, creating...")
            container_config = create_container_config(project, container, password)
            try:
                podman_container = client.containers.create(**container_config)
                start_and_save_container(podman_container, container)
            except Exception as e:
                # TODO Handle cases where the container creation fails because
                # of a missing directory and create it during project creation
                # ideally
                raise PodmanError(f"Failed to start container: {e}")


def stop_container(project: Project, container: Container):
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        try:
            podman_container = client.containers.get(
                f"rstudio_{project.slug}_{container.container_id}"
            )
            podman_container.stop()
        except Exception as e:
            raise PodmanError(f"Failed to stop container: {e}")


def remove_container(project: Project, container: Container):
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        try:
            podman_container = client.containers.get(
                f"rstudio_{project.slug}_{container.container_id}"
            )
            podman_container.remove(force=True)
            print("Container removed.")
        except Exception as e:
            raise PodmanError(f"Failed to remove container: {e}")


def is_container_running(project: Project, container: Container):
    with PodmanClient(base_url=PODMAN_SOCKET) as client:
        try:
            podman_container = client.containers.get(
                f"rstudio_{project.slug}_{container.container_id}"
            )
            return podman_container.status == "running"
        except Exception as e:
            raise PodmanError(f"Failed to get container status: {e}")



# Generated by Django 5.0.4 on 2024-04-20 17:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("api", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="is_active",
            field=models.BooleanField(
                default=False,
                help_text="Indicates whether the project's container is running",
            ),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-21 17:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("api", "0004_project_rstudio_version"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="project",
            name="url",
        ),
        migrations.AddField(
            model_name="project",
            name="local_url",
            field=models.URLField(
                default="",
                help_text="Local URL to access the project's RStudio instance",
                max_length=1024,
            ),
        ),
        migrations.AddField(
            model_name="project",
            name="tailscale_funnel_url",
            field=models.URLField(
                default="",
                help_text="Tailscale Funnel URL to access the project's RStudio instance from the web",
                max_length=1024,
            ),
        ),
        migrations.AddField(
            model_name="project",
            name="tailscale_serve_url",
            field=models.URLField(
                default="",
                help_text="Tailscale Serve URL to access the project's RStudio instance from the tailnet",
                max_length=1024,
            ),
        ),
    ]



# Generated by Django 5.0.4 on 2024-07-01 08:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0020_alter_container_port'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='container',
            name='tailscale_funnel_url',
        ),
        migrations.RemoveField(
            model_name='container',
            name='tailscale_serve_url',
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-20 17:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("api", "0002_project_is_active"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="url",
            field=models.URLField(
                default="http://localhost:8387",
                help_text="URL to access the project's RStudio instance",
                max_length=1024,
            ),
            preserve_default=False,
        ),
    ]



from django.conf import settings
from django.contrib.auth import get_user_model
from django.shortcuts import resolve_url
from django.test.utils import TestContextDecorator
from django_otp import DEVICE_ID_SESSION_KEY
from django_otp.oath import totp
from two_factor.plugins.registry import registry
from two_factor.utils import default_device, totp_digits


class UserMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.login_url = resolve_url(settings.LOGIN_URL)
        cls.User = get_user_model()

    def setUp(self):
        super().setUp()
        self._passwords = {}

    def create_user(self, username="bouke@example.com", password="secret", **kwargs):
        user = self.User.objects.create_user(
            username=username, email=username, password=password, **kwargs
        )
        self._passwords[user] = password
        return user

    def create_superuser(
        self, username="bouke@example.com", password="secret", **kwargs
    ):
        user = self.User.objects.create_superuser(
            username=username, email=username, password=password, **kwargs
        )
        self._passwords[user] = password
        return user

    def login_user(self):
        user = list(self._passwords.keys())[0]
        username = user.get_username()
        assert self.client.login(username=username, password=self._passwords[user])
        if default_device(user):
            session = self.client.session
            session[DEVICE_ID_SESSION_KEY] = default_device(user).persistent_id
            session.save()

    def enable_otp(self, user=None):
        if user is None:
            user = list(self._passwords.keys())[0]
        return user.totpdevice_set.create(name="default")


class method_registry(TestContextDecorator):
    def __init__(self, method_codes):
        self.codes = method_codes
        super().__init__()

    def enable(self):
        # We count on the fact that initially registry._methods is full (default test settings)
        self.old_methods = registry._methods
        registry._methods = [m for m in self.old_methods if m.code in self.codes]

    def disable(self):
        registry._methods = self.old_methods


def totp_str(key):
    return str(totp(key)).zfill(totp_digits())
